@@ -35,76 +35,129 @@ using Beagle.Util;
 
 namespace Beagle.Daemon.GaimLogQueryable {
 
-	[QueryableFlavor (Name="IMLog", Domain=QueryDomain.Local)]
+	[QueryableFlavor (Name="IMLog", Domain=QueryDomain.Local, RequireInotify=false)]
 	public class GaimLogQueryable : LuceneQueryable {
 
 		private static Logger log = Logger.Get ("GaimLogQueryable");
-		private string log_dir;
 
+		private string config_dir, log_dir;
+
+		private int polling_interval_in_seconds = 60;
+		
 		Hashtable watched = new Hashtable ();
 
+		private GaimLogCrawler crawler;
+
 		public GaimLogQueryable () : base ("GaimLogIndex")
 		{
-			string home = Environment.GetEnvironmentVariable ("HOME");
-
-			log_dir = Path.Combine (Path.Combine (home, ".gaim"), "logs");
+			config_dir = Path.Combine (PathFinder.HomeDir, ".gaim");
+			log_dir = Path.Combine (config_dir, "logs");
 		}
 
-		private void StartWorker () 
-		{
-			Inotify.Event += OnInotifyEvent;
+		/////////////////////////////////////////////////
+					
+		private void StartWorker() 
+		{		
+			if (! Directory.Exists (log_dir)) {
+				GLib.Timeout.Add (60000, new GLib.TimeoutHandler (CheckForExistence));
+				return;
+			}
 
-			log.Info ("Scanning IM Logs");
-			Stopwatch timer = new Stopwatch ();
-			timer.Start ();
-			int foundCount = Watch (log_dir);
-			timer.Stop ();
-			log.Info ("Found {0} logs in {1}", foundCount, timer);
+			log.Info ("Starting Gaim log backend");
+
+			Stopwatch stopwatch = new Stopwatch ();
+			stopwatch.Start ();
+
+			if (Inotify.Enabled) {
+				Inotify.Event += OnInotifyEvent;
+				Watch (log_dir);
+			}
+
+			crawler = new GaimLogCrawler (log_dir);
+			Crawl ();
+
+			if (!Inotify.Enabled) {
+				Scheduler.Task task = Scheduler.TaskFromHook (new Scheduler.TaskHook (CrawlHook));
+                                task.Tag = "Crawling ~/.gaim/logs to find new logfiles";
+                                ThisScheduler.Add (task);
+			}
+
+			stopwatch.Stop ();
+
+			log.Info ("Gaim log backend worker thread done in {0}", stopwatch); 
 		}
-
+		
 		public override void Start () 
 		{
-			// FIXME: If ~/.gaim/logs doesn't exist we should set up watches
-			// and wait for it to appear instead of just giving up.
-			if (! Directory.Exists (log_dir))
-				return;
-
 			base.Start ();
-
-			Thread th = new Thread (new ThreadStart (StartWorker));
-			th.Start ();
+			
+			ExceptionHandlingThread.Start (new ThreadStart (StartWorker));
 		}
 
+		/////////////////////////////////////////////////
+
+		private void Crawl ()
+                {
+                        crawler.Crawl ();
+                        foreach (FileInfo file in crawler.Logs) {			    
+                                IndexLog (file.FullName, Scheduler.Priority.Delayed);
+			}
+                }
+
+                private void CrawlHook (Scheduler.Task task)
+                {
+                        Crawl ();
+                        task.Reschedule = true;
+                        task.TriggerTime = DateTime.Now.AddSeconds (polling_interval_in_seconds);
+                }
+
+		/////////////////////////////////////////////////
+
+		// Sets up an Inotify watch on all subdirectories withing ~/.gaim/logs
 		private int Watch (string path)
 		{
 			DirectoryInfo root = new DirectoryInfo (path);
-			if (! root.Exists)
-				return 0;
-
+			
+			if (! root.Exists) {
+				log.Warn ("IM: {0} cannot watch path. It doesn't exist.", path);
+				return 0;	
+			}
+			
 			int file_count = 0;
-
 			Queue queue = new Queue ();
 			queue.Enqueue (root);
 
 			while (queue.Count > 0) {
 				DirectoryInfo dir = queue.Dequeue () as DirectoryInfo;
 				
+				// Setup watches on the present directory.
 				int wd = Inotify.Watch (dir.FullName,
 							Inotify.EventType.CreateSubdir
 							| Inotify.EventType.Modify);
+				
 				watched [wd] = true;
 
-				foreach (FileInfo file in dir.GetFiles ()) {
- 					IndexLog (file.FullName, Scheduler.Priority.Delayed);
-					++file_count;
-				}
-
+				// Add all subdirectories to the queue so their files can be indexed.
 				foreach (DirectoryInfo subdir in dir.GetDirectories ())
 					queue.Enqueue (subdir);
 			}
-
+			
 			return file_count;
 		}
+		
+		/////////////////////////////////////////////////
+
+		private bool CheckForExistence ()
+		{
+			if (!Directory.Exists (log_dir))
+				return true;
+
+			this.Start ();
+
+			return false;
+		}
+
+		/////////////////////////////////////////////////
 
 		private void OnInotifyEvent (int wd,
 					     string path,
@@ -119,16 +172,18 @@ namespace Beagle.Daemon.GaimLogQueryable {
 
 			switch (type) {
 				
-			case Inotify.EventType.CreateSubdir:
-				Watch (full_path);
-				break;
-
-			case Inotify.EventType.Modify:
-				IndexLog (full_path, Scheduler.Priority.Immediate);
-				break;
+				case Inotify.EventType.CreateSubdir:
+					Watch (full_path);
+					break;
+	
+				case Inotify.EventType.Modify:
+					IndexLog (full_path, Scheduler.Priority.Immediate);
+					break;
 			}
 		}
 
+		/////////////////////////////////////////////////
+		
 		private static Indexable ImLogToIndexable (ImLog log)
 		{
 			Indexable indexable = new Indexable (log.Uri);
@@ -163,7 +218,8 @@ namespace Beagle.Daemon.GaimLogQueryable {
 		private void IndexLog (string filename, Scheduler.Priority priority)
 		{
 			FileInfo info = new FileInfo (filename);
-			if (! info.Exists || Driver.IsUpToDate (filename))
+			if (! info.Exists
+			    || this.FileAttributesStore.IsUpToDate (filename))
 				return;
 
 			Scheduler.TaskGroup group;
@@ -179,5 +235,44 @@ namespace Beagle.Daemon.GaimLogQueryable {
 				ThisScheduler.Add (task);
 			}
 		}
+
+		override protected double RelevancyMultiplier (Hit hit)
+		{
+			return HalfLifeMultiplierFromProperty (hit, 0.25,
+							       "fixme:endtime", "fixme:starttime");
+		}
+
+		override public string GetSnippet (QueryBody body, Hit hit)
+		{
+			// FIXME: This does the wrong thing for old-style logs.
+			string file = hit ["fixme:file"];
+			ICollection logs = GaimLog.ScanLog (new FileInfo (file));
+			IEnumerator iter = logs.GetEnumerator ();
+			ImLog log = null;
+			if (iter.MoveNext ())
+				log = iter.Current as ImLog;
+			if (log == null)
+				return null;
+
+			// FIXME: This is very lame, and doesn't do the
+			// right thing w/ stemming, word boundaries, etc.
+			foreach (ImLog.Utterance utt in log.Utterances) {
+				string lower_text = utt.Text.ToLower ();
+				string text = utt.Text;
+				int i = -1;
+				foreach (string query_text in body.Text) {
+					i = lower_text.IndexOf (query_text.ToLower ());
+					if (i >= 0) {
+						text = String.Concat (text.Substring (0, i), "<b>", text.Substring (i, query_text.Length), "</b>", text.Substring (i + query_text.Length));
+						lower_text = text.ToLower ();
+					}
+				}
+				if (i >= 0)
+					return text;
+			}
+
+			// If all else fails, return the log's generic snippet
+			return log.Snippet;
+		}
 	}
 }
