@@ -31,6 +31,7 @@
 
 using System;
 using System.Collections;
+using System.Diagnostics;
 using System.Globalization;
 using System.IO;
 using System.Text;
@@ -49,17 +50,14 @@ using Beagle.Util;
 
 namespace Beagle.Daemon {
 
-	public class LuceneDriver {
+	public class LuceneDriver : IIndexer {
 
 		public delegate bool UriFilter (Uri uri);
+		public delegate double RelevancyMultiplier (Hit hit);
 
+		public event IIndexerChangedHandler ChangedEvent;
 
-		public delegate void ChangedHandler (LuceneDriver source,
-						     ICollection list_of_added_uris,
-						     ICollection list_of_removed_uris);
-
-		public event ChangedHandler ChangedEvent;
-
+		/////////////////////////////////////////////////////
 		
 		// 1: Original
 		// 2: Changed format of timestamp strings
@@ -68,28 +66,57 @@ namespace Beagle.Daemon {
 		// 5: Changed analyzer to support stemming.  Bumped version # to
 		//    force everyone to re-index.
 		// 6: lots of schema changes as part of the general refactoring
-		private const int VERSION = 6;
+		// 7: incremented to force a re-index after our upgrade to lucene 1.4
+		//    (in theory the file formats are compatible, we are seeing 'term
+		//    out of order' exceptions in some cases)
+		private const int VERSION = 7;
 
-		private Hashtable pending_by_uri = new Hashtable ();
+		private string top_dir;
+		private Hashtable pending_by_uri = UriFu.NewHashtable ();
 		private int pending_adds = 0;
 		private int pending_removals = 0;
-		private int adds_since_last_optimization = 0;
-		private int removals_since_last_optimization = 0;
+		private int cycles_since_last_optimization = 0;
 		private bool optimizing = false;
 		private int last_item_count = -1;
 
-		public LuceneDriver (string dir)
+		public LuceneDriver (string index_name)
 		{
-			Setup (dir);
+			Setup (index_name);
 		}
 
-		private void Setup (string dir)
+		public string IndexDirectory {
+			get { return top_dir; }
+		}
+
+		/////////////////////////////////////////////////////
+
+		//
+		// The Lucene Store
+		//
+
+		private Lucene.Net.Store.Directory ourStore = null;
+		private string ourStorePath = null;
+		
+
+		public Lucene.Net.Store.Directory Store {
+			get { return ourStore; }
+		}
+
+		public string StorePath {
+			get { return ourStorePath; }
+		}
+
+		/////////////////////////////////////////////////////
+
+
+		private void Setup (string index_name)
 		{
-			string versionFile = Path.Combine (dir, "version");
-			string fingerprintFile = Path.Combine (dir, "fingerprint");
-			string lockDir = Path.Combine (dir, "Locks");
-			string indexDir = Path.Combine (dir, "Index");
-			string queueDir = Path.Combine (dir, "Queue");
+			top_dir = Path.Combine (PathFinder.StorageDir, index_name); 
+			
+			string versionFile = Path.Combine (top_dir, "version");
+			string fingerprintFile = Path.Combine (top_dir, "fingerprint");
+			string lockDir = Path.Combine (top_dir, "Locks");
+			string indexDir = Path.Combine (top_dir, "Index");
 			string indexTestFile = Path.Combine (indexDir, "segments");
 
 			bool versionExists = File.Exists (versionFile);
@@ -140,20 +167,20 @@ namespace Beagle.Daemon {
 				// Purge and rebuild the index's directory
 				// structure.
 
-				if (Directory.Exists (dir)) {
-					log.Debug ("Purging {0}", dir);
-					Directory.Delete (dir, true);
+				if (Directory.Exists (top_dir)) {
+					log.Debug ("Purging {0}", top_dir);
+					Directory.Delete (top_dir, true);
 				}
 
 				// Create all directories.
-				Directory.CreateDirectory (dir);
+				Directory.CreateDirectory (top_dir);
 				Directory.CreateDirectory (lockDir);
 				Directory.CreateDirectory (indexDir);
 
 				StreamWriter sw;
 
 				// Generate a fingerprint and write it out
-				fingerprint = DateTime.Now.Ticks.ToString ();
+				fingerprint = Guid.NewGuid ().ToString ();
 				sw = new StreamWriter (fingerprintFile, false);
 				sw.WriteLine (fingerprint);
 				sw.Close ();
@@ -167,19 +194,12 @@ namespace Beagle.Daemon {
 			Lucene.Net.Store.FSDirectory store;
 			store = Lucene.Net.Store.FSDirectory.GetDirectory (indexDir, false);
 			store.TempDirectoryName = lockDir;
+			ourStore = store;
+			ourStorePath = indexDir;
 
-			Store = store;
+			//Store = store;
 
-			// Before we start, optimize the index.  We want to do
-			// this every time at start-up to help avoid running
-			// out of file descriptors if the beagled were to
-			// crash or be shut down while in a non-optimized state.
-			// (We do the optimization in a thread since it is I/O bound
-			// and slows down start-up.)
-			if (indexExists) {
-				Thread opt_th = new Thread (new ThreadStart (Optimize));
-				opt_th.Start ();
-			} else {
+			if (!indexExists) {
 				// This creates the index if it doesn't exist
 				IndexWriter writer = new IndexWriter (Store, null, true);
 				writer.Close ();
@@ -210,65 +230,8 @@ namespace Beagle.Daemon {
 			get { return fingerprint; }
 		}
 
-		private IFileAttributesStore fa_store = null;
-
-		public IFileAttributesStore FileAttributesStore {
-			get { return fa_store; }
-			set { fa_store = value; }
-		}
-
-		public bool IsUpToDate (string path)
-		{
-			if (fa_store == null)
-				return false;
-
-			FileAttributes attr = fa_store.Read (path);
-
-			// FIXME: This check is incomplete
-			return attr != null
-				&& attr.Path == path
-				&& FileSystem.GetLastWriteTime (path) <= attr.LastWriteTime;
-		}
-
-		public void AttachTimestamp (string path, DateTime mtime)
-		{
-			if (fa_store == null)
-				return;
-
-			FileAttributes attr = new FileAttributes ();
-
-			// FIXME?: we just assume that there won't be collisions.
-			attr.UniqueId = Guid.NewGuid ().ToString ();
-
-			attr.Path = path;
-			attr.LastWriteTime = mtime;
-			attr.LastIndexedTime = DateTime.Now;
-
-			if (! fa_store.Write (attr)) {
-				log.Warn ("Couldn't store file attributes for {0}", path);
-			}
-		}
-
 		/////////////////////////////////////////////////////
 		
-		//
-		// The Lucene Store
-		//
-
-		private Lucene.Net.Store.Directory ourStore = null;
-
-		public Lucene.Net.Store.Directory Store {
-			get { return ourStore; }
-
-			set {
-				if (ourStore != null)
-					throw new Exception ("Attempt to attach a second store to a LuceneDriver");
-				ourStore = (Lucene.Net.Store.Directory) value;
-			}
-		}
-
-		/////////////////////////////////////////////////////
-
 		//
 		// Public Indexing API
 		//
@@ -327,14 +290,6 @@ namespace Beagle.Daemon {
 					Uri uri = (Uri) entry.Key;
 					Indexable indexable = (Indexable) entry.Value;
 
-					// Filter out indexables with
-					// non-transient file ContentUris that
-					// appear to be up-to-date.
-					if (indexable != null
-					    && indexable.IsNonTransient
-					    && IsUpToDate (indexable.ContentUri.LocalPath))
-						continue;
-					
 					pending_uris.Add (uri);
 					if (indexable != null)
 						pending_indexables.Add (indexable);
@@ -373,20 +328,8 @@ namespace Beagle.Daemon {
 			//Log.Debug ("Step #1: {0} {1} {2}", watch, pending_uris.Count,
 			//	   watch.ElapsedTime / pending_uris.Count);
 
-			
-			// Step #2: Cache non-transient content mtimes
-			Hashtable mtimes = new Hashtable ();
-			foreach (Indexable indexable in pending_indexables) {
-				if (indexable.IsNonTransient) {
-					try {
-						string path = indexable.ContentUri.LocalPath;
-						mtimes [path] = FileSystem.GetLastWriteTime (path);
-					} catch { }
-				}
-			}
 
-
-			// Step #3: Write out the pending adds
+			// Step #2: Write out the pending adds
 			watch.Restart ();
 			IndexWriter writer = null;
 			foreach (Indexable indexable in pending_indexables) {
@@ -413,100 +356,153 @@ namespace Beagle.Daemon {
 			if (writer != null) 
 				writer.Close ();
 			watch.Stop ();
-			//Log.Debug ("Step #3: {0}", watch);
+			//Log.Debug ("Step #2: {0}", watch);
 			
-			// Step #4: Mark added non-transient ContentUri files.
-			watch.Restart ();
-			foreach (Indexable indexable in pending_indexables) {
-				if (indexable.ContentUri.IsFile && ! indexable.DeleteContent) {
-					string path = indexable.ContentUri.LocalPath;
-					if (mtimes.Contains (path))
-						AttachTimestamp (path, (DateTime) mtimes [path]);
-				}
-			}
-			watch.Stop ();
-			//Log.Debug ("Step #4: {0}", watch);
 
+			// Step #3: Fire off an event telling what we just did.
 			if (ChangedEvent != null) {
 				ChangedEvent (this, added_uris, removed_uris);
 			}
 
-			lock (pending_by_uri) {
-				adds_since_last_optimization += add_count;
-				removals_since_last_optimization += removal_count;
-			}
+			lock (pending_by_uri)
+				cycles_since_last_optimization++;
+
+			if (NeedsOptimize)
+				Optimize ();
 		}
 
-		public bool NeedsOptimize {
+		private bool NeedsOptimize {
 			get { 
-				// FIXME: 259 is a totally arbitrary number.
-				return adds_since_last_optimization + removals_since_last_optimization > 259;
+				// FIXME: 19 is a totally arbitrary number.
+				return cycles_since_last_optimization > 19;
 			}
 		}
 
-		public void Optimize ()
+		private void Optimize ()
 		{
 			// If nothing has happened since our last optimization,
 			// do dothing.
 			// If this index is already being optimized, don't
 			// optimize it again.
 			lock (pending_by_uri) {
-				if (optimizing || (adds_since_last_optimization == 0
-						   && removals_since_last_optimization == 0))
+				if (optimizing || cycles_since_last_optimization == 0)
 					return;
 				optimizing = true;
 			}
 
-			Log.Debug ("Optimizing...");
-			IndexWriter writer = new IndexWriter (Store, null, false);
-			writer.Optimize ();
-			writer.Close ();
+			Log.Debug ("Optimizing {0}...", StorePath);
+
+			Stopwatch watch = new Stopwatch ();
+			watch.Start ();
+
+			bool optimize_in_process = true;
+
+#if false
+			if (optimizer_path != null) {
+				try {
+					// Try to optimize in an external process
+					Process p = new Process ();
+					p.StartInfo.UseShellExecute = true;
+					p.StartInfo.FileName = optimizer_path;
+					p.StartInfo.Arguments = StorePath;
+					p.Start ();
+					p.WaitForExit ();
+					optimize_in_process = false;
+				} catch (Exception ex) {
+					Log.Debug ("Caught exception while running beagle-index-optimizer");
+					Log.Debug (ex);
+				}
+			}
+#endif
+
+			if (optimize_in_process) {
+				Log.Debug ("Optimizing in-process.");
+				IndexWriter writer = new IndexWriter (Store, null, false);
+				writer.Optimize ();
+				writer.Close ();
+			}
+
+			watch.Stop ();
+
+			Log.Debug ("Optimization time for {0}: {1}", StorePath, watch);
 
 			lock (pending_by_uri) {
 				optimizing = false;
-				adds_since_last_optimization = 0;
-				removals_since_last_optimization = 0;
+				cycles_since_last_optimization = 0;
 			}
 		}
 
 		/////////////////////////////////////////////////////
 
-		public void DoQuery (QueryBody body, IQueryResult result, ICollection list_of_uris, UriFilter uri_filter)
+		// Returns the lowest matching score before the results are
+		// truncated.
+		public void DoQuery (QueryBody           body,
+				     IQueryResult        result,
+				     ICollection         list_of_uris,
+				     UriFilter           uri_filter,
+				     RelevancyMultiplier relevancy_multiplier)
 		{
+			double t_lucene;
+			double t_assembly;
+
+			Stopwatch sw = new Stopwatch ();
+			sw.Start ();
 			IndexReader reader = IndexReader.Open (Store);
-			if (last_item_count == -1)
-				last_item_count = reader.NumDocs ();
 
 			LNS.Searcher searcher = new LNS.IndexSearcher (reader);
 			LNS.Query query = ToLuceneQuery (body, list_of_uris);
 
 			LNS.Hits hits = searcher.Search (query);
-			int n_hits = hits.Length ();
+			sw.Stop ();
 
+			t_lucene = sw.ElapsedTime;
+
+			//////////////////////////////////////
+
+			sw.Reset ();
+			sw.Start ();
+
+			int n_hits = hits.Length ();
 			if (n_hits == 0)
 				return;
 
-			ArrayList filtered_hits = new ArrayList ();
 			for (int i = 0; i < n_hits; ++i) {
 				Document doc = hits.Doc (i);
+
 				if (uri_filter != null) {
 					Uri uri = UriFromLuceneDoc (doc);
 					if (! uri_filter (uri))
 						continue;
 				}
-				Hit hit = FromLuceneDocToHit (doc, hits.Id (i), hits.Score (i));
-				if (hit != null)
-					filtered_hits.Add (hit);
 
-				if ((i + 1) % 200 == 0) {
-					result.Add (filtered_hits);
-					filtered_hits = new ArrayList ();
+				double score = (double) hits.Score (i);
+
+				if (result.WillReject (score)) {
+					log.Debug ("Terminating DoQuery at {0} of {1} (score={2})", i, n_hits, score);
+					break;
 				}
-			}
-			result.Add (filtered_hits);
 
-			// The call to searcher.Close () closes the IndexReader.
+				Hit hit = FromLuceneDocToHit (doc, hits.Id (i), score);
+
+				if (relevancy_multiplier != null) {
+					double m = relevancy_multiplier (hit);
+					hit.ScoreMultiplier = (float) m;
+				}
+
+				result.Add (hit);
+			}
+
+			sw.Stop ();
+			
+			t_assembly = sw.ElapsedTime;
+
+			//////////////////////////////////////
+
+			// The call to searcher.Close () also closes the IndexReader.
 			searcher.Close ();
+
+			log.Debug ("{0}: n_hits={1} lucene={2:0.00}s  assembly={3:0.00}s",
+				   StorePath, n_hits, t_lucene, t_assembly);
 		}
 
 		public ICollection DoQueryByUri (ICollection list_of_uris)
@@ -659,7 +655,17 @@ namespace Beagle.Daemon {
 		private LNS.Query ToCoreLuceneQuery (QueryBody body, string field)
 		{
 			LNS.BooleanQuery luceneQuery = null;
-			foreach (string text in body.Text) {
+			foreach (string text_orig in body.Text) {
+				string text = text_orig;
+
+				if (text == null || text == "")
+					continue;
+
+				bool minus_sign = false;
+				if (text [0] == '-') {
+					text = text.Substring (1);
+					minus_sign = true;
+				}
 
 				// Use the analyzer to extract the query's tokens.
 				// This code is taken from Lucene's query parser.
@@ -699,7 +705,7 @@ namespace Beagle.Daemon {
 				if (q != null) {
 					if (luceneQuery == null)
 						luceneQuery = new LNS.BooleanQuery ();
-					luceneQuery.Add (q, true, false);
+					luceneQuery.Add (q, !minus_sign, minus_sign);
 				}
 			}
 			return luceneQuery;
@@ -785,7 +791,7 @@ namespace Beagle.Daemon {
 			string uri = doc.Get ("Uri");
 			if (uri == null)
 				throw new Exception ("Got document from Lucene w/o a URI!");
-			return new Uri (uri, true);
+			return UriFu.UriStringToUri (uri);
 		}
 
 		static private void FromLuceneDocToVersioned (Document doc, Versioned versioned)
@@ -802,7 +808,7 @@ namespace Beagle.Daemon {
 
 		}
 
-		private Hit FromLuceneDocToHit (Document doc, int id, float score)
+		private Hit FromLuceneDocToHit (Document doc, int id, double score)
 		{
 			Hit hit = new Hit ();
 
@@ -840,12 +846,122 @@ namespace Beagle.Daemon {
 		// A common, shared analyzer
 		//
 
+		private class BeagleNoiseFilter : TokenFilter {
+			
+			static int total_count = 0;
+			static int noise_count = 0;
+
+			TokenStream token_stream;
+
+			public BeagleNoiseFilter (TokenStream input) : base (input)
+			{
+				token_stream = input;
+			}
+
+			// FIXME: we should add some heuristics that are stricter
+			// but explicitly try to avoid filtering out dates,
+			// phone numbers, etc.
+			private static bool IsNoise (string text)
+			{
+				// Anything really long is almost certainly noise.
+				if (text.Length > 25) 
+					return true;
+
+				// Look at how often we switch between numbers and letters.
+				// Scoring:
+				// <letter> <digit>   1
+				// <digit> <letter>   1
+				// <x> <punct>+ <x>   1
+				// <x> <punct>+ <y>   2
+				const int transitions_cutoff = 4;
+				int last_type = -1, last_non_punct_type = -1, first_type = -1;
+				bool has_letter = false, has_digit = false, has_punctuation = false;
+				int transitions = 0;
+				for (int i = 0; i < text.Length && transitions < transitions_cutoff; ++i) {
+					char c = text [i];
+					int type = -1;
+					if (Char.IsLetter (c)) {
+						type = 1;
+						has_letter = true;
+					} else if (Char.IsDigit (c)) {
+						type = 2;
+						has_digit = true;
+					} else if (Char.IsPunctuation (c)) {
+						type = 3;
+						has_punctuation = true;
+					}
+					
+					if (type != -1) {
+						
+						if (type != last_type) {
+							if (last_type == 3) {
+								if (type != last_non_punct_type)
+									++transitions;
+							} else {
+								++transitions;
+							}
+						}
+
+						if (first_type == -1)
+							first_type = type;
+
+						last_type = type;
+						if (type != 3)
+							last_non_punct_type = type;
+					}
+				}
+
+				// If we make too many transitions, it must be noise.
+				if (transitions >= transitions_cutoff) 
+					return true;
+
+				// If we consist of nothing but digits and punctuation, treat it
+				// as noise if it is too long.
+				if (transitions == 1 && first_type == 1 && text.Length > 10)
+					return true;
+
+				// We are very suspicious of long things that make lots of
+				// transitions
+				if (transitions > 3 && text.Length > 10) 
+					return true;
+
+				// Beware of anything long that contains a little of everything.
+				if (has_letter && has_digit && has_punctuation && text.Length > 10)
+					return true;
+
+				//Logger.Log.Debug ("BeagleNoiseFilter accepted '{0}'", text);
+				return false;
+				
+			}
+
+			public override Lucene.Net.Analysis.Token Next ()
+			{
+				Lucene.Net.Analysis.Token token;
+				while ( (token = token_stream.Next ()) != null) {
+					if (total_count > 0 && total_count % 5000 == 0)
+						Logger.Log.Debug ("BeagleNoiseFilter filtered {0} of {1} ({2:0.0}%)",
+								  noise_count, total_count, 100.0 * noise_count / total_count);
+					++total_count;
+					if (IsNoise (token.TermText ())) {
+						++noise_count;
+						continue;
+					}
+					return token;
+				}
+				return null;
+			}
+		}
+
 		// This is just a standard analyzer combined with the Porter stemmer.
 		// FIXME: This assumes everything being indexed is in English!
 		private class BeagleAnalyzer : StandardAnalyzer {
 			public override TokenStream TokenStream (String fieldName, TextReader reader)
 			{
-				return new PorterStemFilter (base.TokenStream (fieldName, reader));
+				TokenStream outstream = base.TokenStream (fieldName, reader);
+				if (fieldName == "Text" || fieldName == "HotText")
+					outstream = new BeagleNoiseFilter (outstream);
+				outstream = new PorterStemFilter (outstream);
+				return outstream;
 			}
 		}
 
@@ -861,6 +977,19 @@ namespace Beagle.Daemon {
 
 		/////////////////////////////////////////////////////
 
+		//
+		// Access to the Stemmer
+		//
+
+		static PorterStemmer stemmer = new PorterStemmer ();
+
+		static public string Stem (string str)
+		{
+			return stemmer.Stem (str);
+		}
+
+		/////////////////////////////////////////////////////
+
 		//
 		// Helpful little utility functions
 		//
@@ -888,5 +1017,24 @@ namespace Beagle.Daemon {
 				return key.Substring (propPrefix.Length);
 			return null;
 		}
+
+		/////////////////////////////////////////////////////
+
+		// Expose some information for debugging and analytical purposes.
+
+		public void WriteIndexTermFrequencies (TextWriter writer)
+		{
+			IndexReader reader = IndexReader.Open (Store);
+			TermEnum term_enum = reader.Terms ();
+
+			Term term;
+			while (term_enum.Next ()) {
+				term = term_enum.Term ();
+				int freq = term_enum.DocFreq ();
+				writer.WriteLine ("{0} {1} {2}", term.Field (), freq, term.Text ());
+
+			}
+			reader.Close ();
+		}
 	}
 }
