@@ -172,6 +172,22 @@ namespace Beagle.Daemon {
 			return true;
 		}
 
+		// Schedule all non-valid Uris for removal.
+		private bool HitIsValidOrElse (Uri uri)
+		{
+			bool is_valid = HitIsValid (uri);
+
+			if (! is_valid) {
+				// FIXME: There is probably a race here --- what if the hit
+				// becomes valid sometime between calling HitIsValid
+				// and the removal task being executed?
+				Scheduler.Task task = NewRemoveTask (uri);
+				ThisScheduler.Add (task, Scheduler.AddType.DeferToExisting);
+			}
+
+			return is_valid;
+		}
+
 		/////////////////////////////////////////
 
 		virtual protected double RelevancyMultiplier (Hit hit)
@@ -253,7 +269,7 @@ namespace Beagle.Daemon {
 			Driver.DoQuery (body, 
 					query_result,
 					added_uris,
-					new LuceneDriver.UriFilter (HitIsValid),
+					new LuceneDriver.UriFilter (HitIsValidOrElse),
 					new LuceneDriver.RelevancyMultiplier (RelevancyMultiplier));
 		}
 
@@ -268,14 +284,17 @@ namespace Beagle.Daemon {
 			if (reader == null)
 				return null;
 
-			return SnippetFu.GetSnippet (body, reader);
+			string snippet = SnippetFu.GetSnippet (body, reader);
+			reader.Close ();
+
+			return snippet;
 		}
 
 		/////////////////////////////////////////
 
 		public virtual int GetItemCount ()
 		{
-			return Driver.GetItemCount ();
+			return indexer.GetItemCount ();
 		}
 
 		//////////////////////////////////////////////////////////////////////////////////
@@ -416,8 +435,11 @@ namespace Beagle.Daemon {
 			{
 
 				if (indexable != null) {
-					queryable.CacheIndexableInfo (indexable);
-					indexer.Add (indexable);
+					if (! (indexable.Uri.IsFile 
+					       && queryable.FileAttributesStore.IsUpToDate (indexable.Uri.LocalPath))) {
+						queryable.CacheIndexableInfo (indexable);
+						indexer.Add (indexable);
+					}
 				} else if (uri != null) {
 					indexer.Remove (uri);
 				} else if (generator != null) {
@@ -440,6 +462,9 @@ namespace Beagle.Daemon {
 						// This means that the generator didn't have an indexable
 						// to return this time through, but it does not mean that
 						// its processing queue is empty.
+						// FIXME: Shouldn't we just break if generated is null?
+						// Right now we just call GetNextIndexable a bunch of times
+						// when we don't have more work to do.
 						if (generated != null) {
 							queryable.CacheIndexableInfo (generated);
 							indexer.Add (generated);
